import logging
import socket
import ssl
import sys
from http.server import ThreadingHTTPServer, HTTPServer

from tfmock.proxy.proxy2 import ProxyRequestHandler


class ThreadingHTTPSServer(ThreadingHTTPServer):
    address_family = socket.AF_INET6
    daemon_threads = True

    cakey = 'ca.key'
    cacert = 'ca.crt'

    def get_request(self):
        request, client_address = self.socket.accept()
        request = ssl.wrap_socket(request, keyfile=self.cakey, certfile=self.cacert, server_side=True)
        return request, client_address

    def handle_error(self, request, client_address):
        # surpress socket/ssl related errors
        cls, e = sys.exc_info()[:2]
        if cls is socket.error or cls is ssl.SSLError:
            pass
        else:
            return HTTPServer.handle_error(self, request, client_address)


def test(HandlerClass=ProxyRequestHandler, ServerClass=ThreadingHTTPSServer, protocol="HTTP/1.1"):
    if sys.argv[1:]:
        port = int(sys.argv[1])
    else:
        port = 3129
    server_address = ('', port)

    HandlerClass.protocol_version = protocol
    httpd = ServerClass(server_address, HandlerClass)

    sa = httpd.socket.getsockname()
    logging.info("Serving HTTPS Proxy on", sa[0], "port", sa[1], "...")
    httpd.serve_forever()


if __name__ == '__main__':
    test()



import logging
from concurrent.futures import ThreadPoolExecutor

from tfmock.predict import predict
from tfmock.proxy.proxy2 import ProxyRequestHandler, ThreadingHTTPServer

logging.basicConfig(level=logging.INFO, format='%(asctime)s - %(levelname)s: %(message)s')


class SSLStripRequestHandler(ProxyRequestHandler):
    mock_func = None

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def request_handler(self, req, req_body):
        self.close_connection = True

        if self.mock_func is None:
            res = predict(req, req_body)
        else:
            res = self.mock_func(req, req_body)

        if res is not None:
            return res

    def response_handler(self, req, req_body, res, res_body):
        return


class MockServer:
    def __init__(self, mock_func, mock=True, port=8081):
        server_address = ('::1', port)
        SSLStripRequestHandler.protocol_version = "HTTP/1.1"
        SSLStripRequestHandler.mock_func = mock_func
        httpd = ThreadingHTTPServer(server_address, SSLStripRequestHandler)
        sa = httpd.socket.getsockname()
        logging.info(f"Serving HTTP Proxy on {sa[0]} {port} {sa[1]} ...")

        self.mock = mock
        self.httpd = httpd
        self.executor = ThreadPoolExecutor(max_workers=1)
        self.executor.submit(self.start)

    def start(self):
        self.httpd.serve_forever()

    def close(self):
        self.httpd.shutdown()
        self.executor.shutdown(wait=False, cancel_futures=True)



import os.path
import pathlib

from setuptools import setup

root_folder = os.path.dirname(__file__)
long_description = pathlib.Path(os.path.join(root_folder, "readme.md")).read_text()

setup(
    name='tfmock',
    version='0.0.8',
    install_requires=[],
    packages=['tfmock'],
    package_data={
        '': ['config/*', 'proxy/*']
    },
    include_package_data=True,
    long_description=long_description,
    long_description_content_type='text/markdown'
)



# -*- coding: utf-8 -*-
import logging
import subprocess
import sys
import os
import socket
import ssl
import select
import threading
import gzip
import zlib
import time
import json
import re
from http.server import HTTPServer, BaseHTTPRequestHandler
from io import StringIO
from socketserver import ThreadingMixIn
from subprocess import PIPE
from urllib.parse import urlsplit
import http.client


def join_with_script_dir(path):
    return os.path.join(os.path.dirname(os.path.abspath(__file__)), path)


class ThreadingHTTPServer(ThreadingMixIn, HTTPServer):
    address_family = socket.AF_INET6
    daemon_threads = True

    def handle_error(self, request, client_address):
        cls, e = sys.exc_info()[:2]
        if cls is socket.error or cls is ssl.SSLError:
            pass
        else:
            return HTTPServer.handle_error(self, request, client_address)


# noinspection PyBroadException,HttpUrlsUsage,PyPep8Naming
class ProxyRequestHandler(BaseHTTPRequestHandler):
    cakey = join_with_script_dir('ca.key')
    cacert = join_with_script_dir('ca.crt')
    certkey = join_with_script_dir('cert.key')
    certdir = join_with_script_dir('certs/')
    timeout = 5
    lock = threading.Lock()

    def __init__(self, *args, **kwargs):
        self.tls = threading.local()
        self.tls.conns = {}
        self._headers_buffer = []
        os.makedirs(self.certdir, exist_ok=True)
        BaseHTTPRequestHandler.__init__(self, *args, **kwargs)

    def log_error(self, format, *args):
        if isinstance(args[0], socket.timeout):
            return

        self.log_message(format, *args)

    def do_CONNECT(self):
        if os.path.isfile(self.cakey) and os.path.isfile(self.cacert) and os.path.isfile(self.certkey) and os.path.isdir(self.certdir):
            self.connect_intercept()
        else:
            self.connect_relay()

    def connect_intercept(self):
        hostname = self.path.split(':')[0]
        certpath = "%s/%s.crt" % (self.certdir.rstrip('/'), hostname)

        with self.lock:
            if not os.path.isfile(certpath):
                epoch = "%d" % (time.time() * 1000)
                p2 = subprocess.run(["openssl", "req", "-new", "-x509", "-key", self.certkey, "-subj", f"/CN={hostname}", "-days", "365", "-addext", f"subjectAltName=DNS:*.{hostname}, DNS:{hostname}", "-CA", self.cacert, "-CAkey", self.cakey, "-set_serial", epoch, "-out", certpath], stderr=PIPE)
                p2.check_returncode()

        self.wfile.write(f"{self.protocol_version} 200 Connection Established\r\n".encode("utf-8"))
        self.end_headers()

        self.connection = ssl.wrap_socket(self.connection, keyfile=self.certkey, certfile=certpath, server_side=True)
        self.rfile = self.connection.makefile("rb", self.rbufsize)
        self.wfile = self.connection.makefile("wb", self.wbufsize)

        conntype = self.headers.get('Proxy-Connection', '')
        if self.protocol_version == "HTTP/1.1" and conntype.lower() != 'close':
            self.close_connection = 0
        else:
            self.close_connection = 1

    def connect_relay(self):
        address = self.path.split(':', 1)
        address[1] = int(address[1]) or 443
        try:
            s = socket.create_connection(address, timeout=self.timeout)
        except Exception as e:
            self.send_error(502)
            return
        self.send_response(200, 'Connection Established')
        self.end_headers()

        conns = [self.connection, s]
        self.close_connection = 0
        while not self.close_connection:
            rlist, wlist, xlist = select.select(conns, [], conns, self.timeout)
            if xlist or not rlist:
                break
            for r in rlist:
                other = conns[1] if r is conns[0] else conns[0]
                data = r.recv(8192)
                if not data:
                    self.close_connection = 1
                    break
                other.sendall(data)

    def wrap_response(self, status, reason, headers, res_body, res_body_plain):
        encode = "utf-8"
        self.wfile.write(f"{self.protocol_version} {status} {reason}\r\n".encode(encode))
        for k in headers:
            self.wfile.write(f"{k}:{headers.get(k)}\r\n".encode(encode))
        self.end_headers()
        if res_body_plain is not None:
            content_encoding = headers.get('Content-Encoding', 'identity')
            res_body = self.encode_content_body(res_body_plain, content_encoding)
        self.wfile.write(res_body)
        self.wfile.flush()

    def do_GET(self):
        if self.path == 'http://proxy2.test/':
            self.send_cacert()
            return

        req = self
        content_length = int(req.headers.get('Content-Length', 0))
        req_body = self.rfile.read(content_length) if content_length else None

        if req.path[0] == '/':
            if isinstance(self.connection, ssl.SSLSocket):
                req.path = "https://%s%s" % (req.headers['Host'], req.path)
            else:
                req.path = "http://%s%s" % (req.headers['Host'], req.path)

        logging.info(f'connecting {req.command} {req.path}')

        mock_res = self.request_handler(req, req_body)
        if mock_res is not None:
            self.wrap_response(mock_res.get('status'), mock_res.get('reason'), mock_res.get('headers'), mock_res.get('body'), json.dumps(mock_res.get('body')))
            return

        u = urlsplit(req.path)
        scheme, netloc, path = u.scheme, u.netloc, (u.path + '?' + u.query if u.query else u.path)
        assert scheme in ('http', 'https')
        if netloc:
            req.headers['Host'] = netloc
        setattr(req, 'headers', self.filter_headers(req.headers))

        proxy = None
        if os.environ.get('https_proxy') is not None:
            proxy = urlsplit(os.environ.get('https_proxy'))
            netloc = proxy.netloc

        try:
            origin = (scheme, netloc)
            if origin not in self.tls.conns:
                if scheme == 'https':
                    self.tls.conns[origin] = http.client.HTTPSConnection(netloc, timeout=self.timeout)
                else:
                    self.tls.conns[origin] = http.client.HTTPConnection(netloc, timeout=self.timeout)
            conn = self.tls.conns[origin]
            if proxy is not None:
                conn.set_tunnel(u.netloc)
            conn.request(self.command, path, req_body, dict(req.headers))
            res = conn.getresponse()

            version_table = {10: 'HTTP/1.0', 11: 'HTTP/1.1'}
            setattr(res, 'headers', res.msg)
            setattr(res, 'response_version', version_table[res.version])

            # support streaming
            if 'Content-Length' not in res.headers and 'no-store' in res.headers.get('Cache-Control', ''):
                self.response_handler(req, req_body, res, '')
                setattr(res, 'headers', self.filter_headers(res.headers))
                self.relay_streaming(res)
                with self.lock:
                    self.save_handler(req, req_body, res, '')
                return

            res_body = res.read()
        except Exception:
            if origin in self.tls.conns:
                del self.tls.conns[origin]
            self.send_error(502)
            return

        content_encoding = res.headers.get('Content-Encoding', 'identity')
        res_body_plain = None

        try:
            res_body_plain = self.decode_content_body(res_body, content_encoding)
            res_body_modified = self.response_handler(req, req_body, res, res_body_plain)
        except:
            res_body_modified = None

        if res_body_modified is False:
            self.send_error(403)
            return
        elif res_body_modified is not None:
            res_body_plain = res_body_modified
            res_body = self.encode_content_body(res_body_plain, content_encoding)
            res.headers['Content-Length'] = str(len(res_body))

        setattr(res, 'headers', self.filter_headers(res.headers))

        self.wrap_response(res.status, res.reason, res.headers, res_body, res_body_plain)

    def relay_streaming(self, res):
        self.wfile.write(f"{self.protocol_version} {res.status} {res.reason}\r\n".encode("utf-8"))
        for line in res.headers.headers:
            self.wfile.write(line)
        self.end_headers()
        try:
            while True:
                chunk = res.read(8192)
                if not chunk:
                    break
                self.wfile.write(chunk)
            self.wfile.flush()
        except socket.error:
            # connection closed by client
            pass

    do_HEAD = do_GET
    do_POST = do_GET
    do_PUT = do_GET
    do_DELETE = do_GET
    do_OPTIONS = do_GET
    do_PATCH = do_GET

    def filter_headers(self, headers):
        # http://tools.ietf.org/html/rfc2616#section-13.5.1
        hop_by_hop = ('connection', 'keep-alive', 'proxy-authenticate', 'proxy-authorization', 'te', 'trailers', 'transfer-encoding', 'upgrade')
        for k in hop_by_hop:
            del headers[k]

        # accept only supported encodings
        if 'Accept-Encoding' in headers:
            ae = headers['Accept-Encoding']
            filtered_encodings = [x for x in re.split(r',\s*', ae) if x in ('identity', 'gzip', 'x-gzip', 'deflate')]
            headers['Accept-Encoding'] = ', '.join(filtered_encodings)

        return headers

    def encode_content_body(self, text, encoding):
        if encoding == 'identity':
            if isinstance(text, bytes):
                return text
            return text.encode("utf-8")
        elif encoding in ('gzip', 'x-gzip'):
            io = StringIO()
            with gzip.GzipFile(fileobj=io, mode='wb') as f:
                f.write(text)
            data = io.getvalue().encode("utf-8")
        elif encoding == 'deflate':
            data = zlib.compress(text)
        else:
            raise Exception("Unknown Content-Encoding: %s" % encoding)
        return data

    def decode_content_body(self, data, encoding):
        if encoding == 'identity':
            text = data
        elif encoding in ('gzip', 'x-gzip'):
            if isinstance(data, bytes):
                io = StringIO()
            else:
                io = StringIO(data)
            with gzip.GzipFile(fileobj=io) as f:
                if isinstance(data, bytes):
                    f.write(data)
                else:
                    text = f.read()
            if isinstance(data, bytes):
                text = io.getvalue()

        elif encoding == 'deflate':
            try:
                text = zlib.decompress(data)
            except zlib.error:
                text = zlib.decompress(data, -zlib.MAX_WBITS)
        else:
            raise Exception("Unknown Content-Encoding: %s" % encoding)
        return text

    def send_cacert(self):
        with open(self.cacert, 'rb') as f:
            data = f.read()

        self.wfile.write("%s %d %s\r\n" % (self.protocol_version, 200, 'OK'))
        self.send_header('Content-Type', 'application/x-x509-ca-cert')
        self.send_header('Content-Length', len(data))
        self.send_header('Connection', 'close')
        self.end_headers()
        self.wfile.write(data)

    def request_handler(self, req, req_body) -> dict:
        pass

    def response_handler(self, req, req_body, res, res_body) -> str:
        pass

    def save_handler(self, req, req_body, res, res_body):
        pass
